#include <stdio.h>
#include <stdlib.h>
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>

#ifdef dDOUBLE
#define dsDrawCapsule dsDrawCapsuleD
#define dsDrawBox     dsDrawBoxD
#define dsDrawLine    dsDrawLineD
#endif

#define LINK_NUM 3
#define JT_NUM   3
#define LEG_NUM  4

dWorldID      world;
dSpaceID      space;
dGeomID       ground;
dJointGroupID contactgroup;
dsFunctions   fn;

typedef struct {
  dBodyID  body;
  dGeomID  geom;
  dJointID joint;
  dReal    m,lx, ly, lz, x,y,z;
  double THETA;
} MyLink, *qMyLink;

MyLink leg[LEG_NUM][LINK_NUM],torso;

dReal Torso_Pos[3] = {0, 0, 0.8};
dReal Link_Up_Size[3] = {0.05, 0.05, 0.2};
dReal Link_middle_size[3] = {0.05, 0.05, 0.2};
dReal Link_Down_size[3] = {0.1, 0.05, 0.05};
dReal Link_size_x[LINK_NUM] = { Link_Up_Size[0], Link_middle_size[0], Link_Down_size[0]};
dReal Link_size_y[LINK_NUM] = { Link_Up_Size[1], Link_middle_size[1], Link_Down_size[1]};
dReal Link_size_z[LINK_NUM] = { Link_Up_Size[2], Link_middle_size[2], Link_Down_size[2]};
dReal Torso_Size[3] = {0.6, 0.4, 0.2};
dReal cx1 = (Torso_Size[0] - Link_Up_Size[0]) / 2;
dReal cy1 = (Torso_Size[1] + Link_Up_Size[1]) / 2;

dReal  c_x[LEG_NUM][LINK_NUM] = {{ cx1, cx1, cx1},{-cx1,-cx1,-cx1},
                                 {-cx1,-cx1,-cx1},{ cx1, cx1, cx1}};
dReal  c_y[LEG_NUM][LINK_NUM] = {{ cy1, cy1, cy1},{ cy1, cy1, cy1},
                                 {-cy1,-cy1,-cy1},{-cy1,-cy1,-cy1}};
dReal  c_z[LEG_NUM][LINK_NUM] =  {{0, -Link_Up_Size[2], -Link_Up_Size[2] - Link_middle_size[2]},{0, -Link_Up_Size[2], -Link_Up_Size[2] - Link_middle_size[2]},
                                {0, -Link_Up_Size[2], -Link_Up_Size[2] - Link_middle_size[2]},{0, -Link_Up_Size[2], -Link_Up_Size[2] - Link_middle_size[2]}};

dReal  p_x[LEG_NUM][LINK_NUM] = {{ cx1, cx1, cx1 + Link_middle_size[0] / 2},{-cx1,-cx1,-cx1 + Link_middle_size[0] / 2},
                                 {-cx1,-cx1,-cx1 + Link_middle_size[0] / 2},{ cx1, cx1, cx1 + Link_middle_size[0] / 2}};
dReal  p_y[LEG_NUM][LINK_NUM] = {{ cy1, cy1, cy1},{ cy1, cy1, cy1},
                                 {-cy1,-cy1,-cy1},{-cy1,-cy1,-cy1}};
dReal  p_z[LEG_NUM][LINK_NUM] =  {{-Link_Up_Size[2] / 2, -Link_Up_Size[2] -  Link_middle_size[2] / 2,  -Link_Up_Size[2] - Link_middle_size[2] - Link_Down_size[2] / 2},{-Link_Up_Size[2] / 2, -Link_Up_Size[2] -  Link_middle_size[2] / 2,  -Link_Up_Size[2] - Link_middle_size[2] - Link_Down_size[2] / 2},
                                {-Link_Up_Size[2] / 2, -Link_Up_Size[2] -  Link_middle_size[2] / 2,  -Link_Up_Size[2] - Link_middle_size[2] - Link_Down_size[2] / 2},{-Link_Up_Size[2] / 2, -Link_Up_Size[2] -  Link_middle_size[2] / 2,  -Link_Up_Size[2] - Link_middle_size[2] - Link_Down_size[2] / 2}};
dReal Link_mass[3] = {0.5, 0.5, 0.2};
dReal Torso_mass = 2.0;

void  makeRobot()
{
    dMass mass;
    torso.body  = dBodyCreate(world);
    dMassSetZero(&mass);
    dMassSetBoxTotal(&mass,Torso_mass, Torso_Size[0], Torso_Size[1], Torso_Size[2]);
    dBodySetMass(torso.body,&mass);
    torso.geom = dCreateBox(space,Torso_Size[0], Torso_Size[1], Torso_Size[2]);
    dGeomSetBody(torso.geom, torso.body);
    dBodySetPosition(torso.body, Torso_Pos[0], Torso_Pos[1], Torso_Pos[2]);


    for (int i = 0; i < LEG_NUM; i++) {
        for (int j = 0; j < LINK_NUM; j++) {
        leg[i][j].body = dBodyCreate(world);
        dBodySetPosition(leg[i][j].body, Torso_Pos[0]+p_x[i][j], Torso_Pos[1]+p_y[i][j], Torso_Pos[2]+p_z[i][j]);
        dMassSetZero(&mass);
        dMassSetBoxTotal(&mass,Link_mass[j], Link_size_x[j], Link_size_y[j], Link_size_z[j]);
        dBodySetMass(leg[i][j].body, &mass);
        leg[i][j].geom = dCreateBox(space,Link_size_x[j], Link_size_y[j], Link_size_z[j]);
        dGeomSetBody(leg[i][j].geom,leg[i][j].body);
        }
    }
    for (int i = 0; i < LEG_NUM; i++) {
        for (int j = 0; j < LINK_NUM; j++) {
            leg[i][j].joint = dJointCreateHinge(world, 0);
            if (j == 0)
                dJointAttach(leg[i][j].joint, torso.body, leg[i][j].body);
            else
                dJointAttach(leg[i][j].joint, leg[i][j-1].body, leg[i][j].body);
            dJointSetHingeAnchor(leg[i][j].joint, Torso_Pos[0]+c_x[i][j], Torso_Pos[1]+c_y[i][j],Torso_Pos[2]+c_z[i][j]);
            dJointSetHingeAxis(leg[i][j].joint, 0, 1, 0);
        }
    }
}

void drawRobot()
{
    dReal r,length;
    dVector3 sides;

    dsSetColor(0.8,0.3,0.3);
    dGeomBoxGetLengths(torso.geom,sides);
    dsDrawBox(dBodyGetPosition(torso.body),
	 					 dBodyGetRotation(torso.body),sides);

    for (int i = 0; i < LEG_NUM; i++) {
        for (int j = 0; j < LINK_NUM; j++ ) {
            if(j == 0)dsSetColor(0.8,0.3,0.8);
            if(j == 1)dsSetColor(0.3,0.8,0.3);
            if(j == 2)dsSetColor(0.3,0.3,0.8);
            dGeomBoxGetLengths(leg[i][j].geom, sides);
    dsDrawBox(dBodyGetPosition(leg[i][j].body),
	 					 dBodyGetRotation(leg[i][j].body),sides);
        }
    }
}

static void nearCallback(void *data, dGeomID o1, dGeomID o2) {
  dBodyID b1 = dGeomGetBody(o1), b2 = dGeomGetBody(o2);
  if (b1 && b2 && dAreConnectedExcluding(b1,b2,dJointTypeContact)) return;

  static const int N = 20;
  dContact contact[N];
  int n = dCollide(o1,o2,N,&contact[0].geom,sizeof(dContact));
  if (n > 0) {
    for (int i=0; i<n; i++) {
      contact[i].surface.mode = dContactSoftERP | dContactSoftCFM;
      contact[i].surface.mu   = dInfinity;
      contact[i].surface.soft_erp = 0.9;
      contact[i].surface.soft_cfm = 1e-5;
	    dJointID c = dJointCreateContact(world,contactgroup,&contact[i]);
      dJointAttach(c,b1,b2);
    }
  }
}

void Pcontrol(qMyLink link)
{
  dReal kp = 2.0, fMax = 100.0;

  for (int i = 0; i < LEG_NUM; i++) {
    for (int j = 0; j < LINK_NUM; j++) {
      dReal tmp = dJointGetHingeAngle(link->joint);
      dReal diff = link->THETA - tmp;
      dReal u = kp * diff;
      dJointSetHingeParam(link->joint,  dParamVel, u);
      dJointSetHingeParam(link->joint, dParamFMax, fMax);
    }
  }
}

void walk()
{
  static int t = 0, steps = 0;
  int interval = 50;

  if ((steps++ % interval)==0){
	  t++;
	}
  else {
    for (int leg_no = 0; leg_no < LEG_NUM; leg_no++) {
      for (int joint_no = 0; joint_no < JT_NUM; joint_no++) {
        leg[leg_no][joint_no].THETA = 0;
        Pcontrol(&leg[leg_no][joint_no]);
      }
    }
  }
}

void simLoop(int pause)
{
  if (!pause) {
    walk();
    dSpaceCollide(space,0,&nearCallback);
    dWorldStep(world, 0.01);
    dJointGroupEmpty(contactgroup);
  }
  drawRobot();
}

void start()
{
  float xyz[3] = {  1.0f,  -1.2f, 0.5f};
  float hpr[3] = {121.0f, -10.0f, 0.0f};
  dsSetViewpoint(xyz,hpr);
  dsSetSphereQuality(3);
  dsSetCapsuleQuality(6);
}

void setDrawStuff() {
  fn.version = DS_VERSION;
  fn.start   = &start;
  fn.step    = &simLoop;
  fn.command = NULL;
  fn.path_to_textures = "../../drawstuff/textures";
}

int main(int argc, char *argv[])
{
  dInitODE();
  setDrawStuff();
  world        = dWorldCreate();
  space        = dHashSpaceCreate(0);
  contactgroup = dJointGroupCreate(0);
  ground       = dCreatePlane(space,0,0,1,0);
  dWorldSetGravity(world, 0, 0, -9.8);
  dWorldSetCFM(world, 1e-3);
  dWorldSetERP(world, 0.9);
  makeRobot();
  dsSimulationLoop(argc,argv,800,480,&fn);
  dSpaceDestroy(space);
  dWorldDestroy(world);
  dCloseODE();
  return 0;
}


